# file: vdb.lua #

# --------------------------------------------------------------------------------------------------

#   In case of error rpmbuild prints source code of the failed chunk. Lets split Lua code to
#   smaller chunks to reduce output in case of problems.

# --------------------------------------------------------------------------------------------------

%{lua: -- Functions I frequently use in rpm spec files.

if not _vdb_loaded_ then

_subpackages_ = {}

-- Escape % characters, return string with all % characters doubled.
function escape( str )
    return string.gsub( str, "%%", "%%%%" )
end

-- Quote string for using in shell scripts.
function quote( str )
    return "'" .. string.gsub( str, "'", "'\\''" ) .. "'"
end

function shrink( str )
    str = string.gsub( str, "^%s+", "" )
    str = string.gsub( str, "%s+", " " )
    str = string.gsub( str, "%s+$", "" )
    return str
end

-- Perl-like keys. Returns a list of hash keys.
function keys( hash )
    local keys = {}
    for key in pairs( hash ) do
        push( keys, key )
    end
    table.sort( keys, function ( a, b ) return tostring( a ) < tostring( b ) end )
    return keys
end

-- Perl-like grep.
function grep( cond, list )
    local result = {}
    for i, value in ipairs( list ) do
        if cond( value ) then
            push( result, value )
        end
    end
    return result
end

-- Perl-like push: append elements to an array.
function push( array, ... )
    local items = { ... }
    for i, item in ipairs( items ) do
        table.insert( array, item )
    end
end

-- Perl-like pop: remove the last element from an array and return it.
function pop( array )
    local item
    if array > 0 then
        item = table.remove( array )
    end
    return item
end

-- Perl-like shift. Returns the first element of array, removing it from the array.
function shift( array )
    local item
    if #array > 0 then
        item = table.remove( array, 1 )
    end
    return item
end

-- Perl-like map: apply func to the each element of the list.
function map( func, list )
    local result = {}
    for i, v in ipairs( list ) do
        push( result, func( v ) )
    end
    return result
end

-- Perl-like join: concatenate all list elements.
function join( sep, list )
    return table.concat( list, sep )
end

-- Perl-like split: split string into items.
function split( sep, str, limit )
    local list = {}
    local len = string.len( str )
    local pos = 1
    if limit == nil then
        limit = 0
    end
    while pos <= len do
        local start, finish -- sep position
        local item
        if limit <= 0 or ( limit > 0 and #list + 1 < limit ) then
            start, finish = string.find( str, sep, pos )
        end
        if start == nil then
            item = string.sub( str, pos )
        else
            item = string.sub( str, pos, start - 1 )
        end
        push( list, item )
        if finish == nil then
            pos = len + 1
        else
            pos = finish + 1
        end
    end
    -- If limit is zero, drop trailing empty items:
    if limit == 0 then
        while #list > 0 and list[ #list ] == "" do
            pop( list )
        end
    end
    return list
end

-- Perl-like uc: convert string to upper case.
uc = string.upper

-- Perl-like lc: convert string to lower case.
lc = string.lower

-- Perl-like sprintf.
sprintf = string.format

-- Perl-like printf.
function printf( format, ... )
    print( sprintf( format, ... ) )
end

-- Expand rpm macros in string.
expand = rpm.expand

-- Define rpm macro.
function set( name, value )
    -- If value is empty, rpm will complain on empty macro value.
    -- Let's use `%{nil}` in such a case to avoid warning.
    if value == nil or value == "" then
        value = "%{nil}"
    end
    local l = name .. " " .. value
    dbg( "%%define %s", l )
    rpm.define( l )
    return value
end

-- Undefine rpm macro.
function undef( name )
    dbg( "%%undefine %s", name )
    rpm.undefine( name )
end

-- Returns true if rpm macro is defined, and false otherwise.
function defined( name )
    return expand( "%{?" .. escape( name ) .. ":1}" ) == "1"
end

-- Get value of rpm macro. If macro is not defined, nil is returned.
function get( name, default )
    if defined( name ) then
        return expand( "%{" .. escape( name ) .. "}" )
    end
end

-- Generate spec code. If the first parameter is table, it is interpreted as set of macros, which
-- will be temporary defined.
function spec( ... )
    local lines = { ... }
    -- See if macros are present.
    local macros = {}
    if #lines > 0 and type( lines[ 1 ] ) == "table" then
        macros = shift( lines )
    end
    -- Save values of macros we are going to set:
    local orig = {}
    for name in pairs( macros ) do
        orig[ name ] = get( name )
    end
    -- Set macros:
    for name, value in pairs( macros ) do
        set( name, value )
    end
    -- Generate spec code:
    for i, line in ipairs( lines ) do
        local l = expand( line )
        dbg( "%s", l )
        if #lines == 1 then
            print( l )
        else
            print( l .. "\n" )
        end
    end
    -- Restore original macro values:
    for name in pairs( orig ) do
        undef( name )
    end
end

local function _say( prefix, format, ... )
    local message = sprintf( format, ... )
    io.stderr:write( prefix .. string.gsub( message, "\n", "\n" .. prefix ) .. "\n" )
end

-- Print a message to stderr if debug enabled.
function dbg( format, ... )
    if _vdb_debug_ then
        _say( "# ", format, ... )
    end
end

-- Print a message to stderr.
function say( format, ... )
    _say( "", format, ... )
end

function warn( format, ... )
    _say( "warning: ", format, ... )
end

-- Print a error to stderr.
-- Building will be stopped in the nearest %prologue.
function die( format, ... )
    _say( "error: ", format, ... )
    error( "oops" )
end

-- Returns true if file exists.
function exist( file )
    return not not posix.stat( file )
end

-- Read file.
-- Read first `limit` lines of file. If limit is zero or nil, read all the lines, but drop empty
-- lines at the end. If `limit` is negative, read all the lines.
function slurp( path, limit )
    local lines = {}
    if limit == nil then
        limit = 0
    end
    for line in io.lines( path ) do
        push( lines, line )
        if limit > 0 and #lines >= limit then
            break
        end
    end
    if limit == 0 then
        while #lines > 0 and lines[ #lines ] == "" do
            pop( lines )
        end
    end
    return lines
end

-- Write file.
function spurt( path, ... )
    say( "+ > %s", path )
    local file, err = io.open( path .. ".tmp", "w" )
    if file == nil then
        die( "%s", err )
    end
    local lines = { ... }
    for i, line in ipairs( lines ) do
        file, err = file:write( line .. "\n" )
        if file == nil then
            die( "%s", err )
        end
    end
    -- TODO: Does close report errors?
    file:close()
    local ok, err = os.rename( path .. ".tmp", path )
    if not ok then
        die( "%s", err )
    end
end

-- Rename file or directory.
function rename( old, new )
    -- `+` mimics `rpmbuild` output.
    say( "+ mv %s %s", old, new )
    local ok, error = os.rename( old, new )
    if not ok then
        die( "mv: %s", error )
    end
    return ok
end

-- Execute shell command.
function shell( command )
    -- `+` mimics `rpmbuild` output.
    say( "+ %s", command )
    local ok, status, code = os.execute( command )
    if not ok then
        if status == "exit" then
            die( "command exited with status %d", code )
        else
            die( "command died with signal %d", code )
        end
    end
    return ok
end

end

}

%{lua:

if not _vdb_loaded then

function banner( phase )
    return ": ---------- " .. phase .. " ---------- : +"
end

-- Prologue to execute in the beginnning of every scriptlet:
local prologue = "set -e; [ -n \"$BASH_VERSION\" ] && set -o pipefail;"
set( "prologue", prologue )

function prep()
    -- Someone can run `rpmbuild -bb name.spec` and build rpm files. However, exact name of built
    -- rpm is not known to the caller because name contains version, release, and architecture --
    -- these parameters are encoded into spec and not easily available outside of the spec file.
    -- Let us write them out, so the caller can source the file and construct rpm name, e. g.
    -- "$NAME-$VERSION-$RELEASE.$ARCH.rpm".
    local name = os.getenv( "SPEC_RC_FILE" )
    if name then
        local arch = get( "buildarch" )
        if arch == nil then
            arch = get( "_arch" )
        end
        spurt(
            name,
            "NAME="    .. quote( get( "name"       ) ),
            "VERSION=" .. quote( get( "version"    ) ),
            "RELEASE=" .. quote( get( "release"    ) ),
            "ARCH="    .. quote( arch                ),
            "RPMS="    .. quote( get( "_rpmdir"    ) ),
            "SRPMS="   .. quote( get( "_srcrpmdir" ) ),
            "SUBPACKAGES=(" .. join( " ", map( quote, _subpackages_ ) ) .. ")"
        )
    end
    spec( banner( "PREP" ), prologue )
end

function build()
    spec( banner( "BUILD" ), prologue )
end

function check()
    spec( banner( "CHECK" ), prologue )
end

function install()
    spec( banner( "INSTALL" ), prologue )
end

function clean()
    spec( banner( "CLEAN" ), prologue )
end

end

}

# --------------------------------------------------------------------------------------------------

%{lua:

if not _vdb_loaded_ then

-- Redefine %prep, %build, %check and %install macros:
-- Let's add prologue to the every build-time scriptlet.
-- The idea is great but Fedora/RedHat rpmbuild macros also use this trick.
-- So let's disable it for now.
local phases = { "prep" }

--~ local phases = { "prep", "build", "check", "install", "clean" }
for i, phase in ipairs( phases ) do
    if defined( phase ) then
        warn( "'%s' macro already defined", phase )
    else
        set( phase, "%%" .. phase .. " \\\n" .. "%{lua: " .. phase .. "();  }\n" )
    end
end

end

}

# --------------------------------------------------------------------------------------------------

%{lua: -- Let's add prologue to the every install-time scriptlet.

if not _vdb_loaded_ then

-- Install-time scriptlets:
local phases = {
    "pretrans",
    "pre",
    "post",
    "triggerin",
    "triggerun",
    "preun",
    "postun",
    "triggerpostun",
    "posttrans",
}

for i, phase in ipairs( phases ) do
    if defined( phase ) then
        warn( "'%s' macro already defined", phase )
    else
        -- If -p option present present, do not add prologue.
        set( phase .. "(p:)", "%%" .. phase .. " %** \\\n" .. "%{!-p:%prologue}\n" )
    end
end

end

}

# --------------------------------------------------------------------------------------------------

%{lua:

if not _vdb_loaded_ then

-- Play the same trick to collect all subpackage names:
if defined( "package" ) then
    die( "'%s' macro already defined", "package" )
end
set(
    -- %package macro may have -n option. If -n is present, it defines full name of the subpackage.
    -- Oterwise subpackage name is constructed from package name and the first macro argument.
    "package(n:)",
    "%%package %** \\\n" .. "%{lua: push( _subpackages_, expand( \"%{-n*}%{!-n:%name-%1}\" ) ) }\n"
)

end

}

%{lua:

if not _vdb_loaded_ then

--[[
    Usage:

        %Global NAME VALUE

    Defines global with specified NAME and VALUE, as well as lowercase version (both name and
    value) of the global.

    Example:

        %Global Font Hack
            # equivalent for
            #   %global Font Hack
            #   %global font hack
]]--
set(
    "Global()",
    "%{lua: spec( " ..
        "\"%global \" .. get( 1 ) .. \" \" .. get( 2 ), " ..
        "\"%global \" .. lc( get( 1 ) ) .. \" \" .. lc( get( 2 ) ) " ..
        ") }"
)

set( "_licensedir", "%{_datadir}/licenses" )
set( "_fontdir",    "%{_datadir}/fonts" )

end

}

%{lua:

_vdb_loaded_ = 1

}

# --------------------------------------------------------------------------------------------------

# end of file #
